// git-remote-k8s - a git remote helper to use PVCs as a remote
// Copyright (C) 2023  James Andariese
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

use crate::{*,errors::*};

#[derive(ClapParser, Debug, Clone)]
/// git-remote-k8s
///
/// This should usually be run by git.  You can set git up to use it
/// by running git remote add k k8s://default/ns/repo.
/// 
/// See <https://crates.io/crates/git-remote-k8s> for more info.
pub struct Config {
    #[arg(
        short,
        long,
        env = "GIT_REMOTE_K8S_IMAGE",
        default_value = "alpine/git:latest"
    )]
    /// Docker image used for git Jobs
    pub image: String,

    #[arg(short,long,env="GIT_REMOTE_K8S_TIMEOUT")]
    pub timeout: Option<humantime::Duration>,

    #[arg(long("setup-timeout"),env = "GIT_REMOTE_K8S_SETUP_TIMEOUT")]
    pub setup_timeout: Option<humantime::Duration>,

    #[arg(long("transfer-timeout"), env = "GIT_REMOTE_K8S_TRANSFER_TIMEOUT")]
    pub transfer_timeout: Option<humantime::Duration>,

    #[arg(index = 1)]
    /// remote name
    pub remote_name: String,

    #[arg(index = 2)]
    /// remote URL
    pub remote_url: String,

    #[arg(
        short,
        long,
        env = "GIT_REMOTE_K8S_DEBUG",
        action=ArgAction::Count,
    )]
    /// verbosity, may be specified more than once
    pub verbose: u8,

    #[arg(
        long("storage-class-name"),
        env = "GIT_REMOTE_K8S_PVC_STORAGECLASSNAME"
    )]
    /// storageClassName to use for the backing PVC _if it is created_.
    ///
    /// this will be used to create a new PVC but if a PVC already
    /// exists by the requested name, it will be used.
    pub storageclass: Option<String>,

    #[arg(
        short('s'),
        long("initial-volume-size"),
        default_value("1Gi"),
        env = "GIT_REMOTE_K8S_INITIAL_VOLUME_SIZE"
    )]
    pub initial_size: String,
}


impl Config {
    /// parse and validate a k8s remote pvc short-URL into a triple of Strings of the form (context, namespace, pvc)
    ///
    /// this utilizes a regex instead of url::Url to ensure that it returns sensible errors
    // TODO: find a way to memoize this cleanly.  probably give it access to a memoizing context from AppContext.
    pub fn parse_and_validate(&self) -> Result<(String, String, String)> {
        let caps = REMOTE_PATTERN
            .captures(&self.remote_url)
            .ok_or(ConfigError::RemoteInvalid)?;
        let scheme = if caps.name("scheme_prefix").is_none() {
            SCHEME
        } else {
            caps.name("scheme")
                .ok_or(ConfigError::RemoteInvalidScheme)?
                .as_str()
        };
        if scheme != SCHEME {
            bail!(ConfigError::RemoteInvalidScheme);
        }
        let kctx = caps
            .name("context")
            .ok_or(ConfigError::RemoteNoContext)?
            .as_str();
        let ns = caps
            .name("namespace")
            .ok_or(ConfigError::RemoteNoNamespace)?
            .as_str();
        let pvc = caps.name("pvc").ok_or(ConfigError::RemoteNoPVC)?.as_str();
        // regex::Regex::find(REMOTE_PATTERN);
        if kctx == "" {
            bail!(ConfigError::RemoteNoContext);
        };
        if ns == "" {
            bail!(ConfigError::RemoteNoNamespace);
        };
        if pvc == "" {
            bail!(ConfigError::RemoteNoPVC);
        };
        if let Some(trailing) = caps.name("trailing") {
            if trailing.as_str() != "" {
                bail!(ConfigError::RemoteTrailingElements);
            }
        }
        Ok((kctx.to_owned(), ns.to_owned(), pvc.to_owned()))
    }

    pub fn get_remote_context(&self) -> Result<String> {
        let (r, _, _) = self.parse_and_validate()?;
        Ok(r)
    }

    pub fn get_remote_namespace(&self) -> Result<String> {
        let (_, r, _) = self.parse_and_validate()?;
        Ok(r)
    }

    pub fn get_remote_pvc(&self) -> Result<String> {
        let (_, _, r) = self.parse_and_validate()?;
        Ok(r)
    }
}