// git-remote-k8s - a git remote helper to use PVCs as a remote
// Copyright (C) 2023  James Andariese
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Affero General Public License for more details.
//
// You should have received a copy of the GNU Affero General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

// TODO: there is a stupid panic here that will be annoying to fix.
//       it happens when there is a timeout that cancels the tasks
//       listening to these sends while this one is sending resulting
//       in a totally ignorable error that I don't know what to do
//       with right now. 
// 2023-11-01T05:02:05.009858Z  INFO git_remote_k8s: waiting for worker pod to come online
// 2023-11-01T05:02:05.018697Z ERROR git_remote_k8s: deadline has elapsed
// 2023-11-01T05:02:05.019458Z ERROR kube_client::client::builder: failed with error channel closed
// thread 'tokio-runtime-worker' panicked at src/pod_utils.rs:118:36:
// couldn't send to channel: SendError { .. }
// note: run with `RUST_BACKTRACE=1` environment variable to display a backtrace
// thread 'tokio-runtime-worker' panicked at src/pod_utils.rs:131:14:
// couldn't send on channel: SendError { .. }


use crate::*;

pub trait PodExt {
    fn label_selectors(&self) -> Vec<String>;
    fn label_selector(&self) -> String;
    fn field_selectors(&self) -> Result<Vec<String>>;
    fn field_selector(&self) -> Result<String>;
}

impl PodExt for Pod {
    fn label_selectors(&self) -> Vec<String> {
        let l = self.labels();
        let selectors = Vec::with_capacity(l.len());
        for (k, v) in l.iter() {
            format!("{}={}", k, v);
        }
        selectors
    }

    fn label_selector(&self) -> String {
        self.label_selectors().join(",").to_owned()
    }

    fn field_selectors(&self) -> Result<Vec<String>> {
        Ok(vec![
            format!(
                "metadata.name={}",
                self.meta()
                    .name
                    .as_ref()
                    .ok_or(ApplicationError::PodNoName)?
            ),
            format!(
                "metadata.namespace={}",
                self.meta()
                    .namespace
                    .as_ref()
                    .ok_or(ApplicationError::PodNoNamespace)?
            ),
        ])
    }

    fn field_selector(&self) -> Result<String> {
        Ok(self.field_selectors()?.join(",").to_owned())
    }
}

pub async fn wait_for_pod_running_watch(pods: &Api<Pod>, pod: Pod) -> Result<()> {
    let wp = default!(WatchParams{
        field_selector: Some(pod.field_selector()?),
    });
    let mut stream = pods.watch(&wp, "0").await?.boxed();
    while let Some(status) = stream.try_next().await? {
        match status {
            WatchEvent::Modified(o) => {
                let s = o.status.as_ref().expect("status exists on pod");
                if s.phase.clone().unwrap_or_default() == "Running" {
                    info!("Ready to attach to {}", o.name_any());
                    break;
                }
                if s.phase.clone().unwrap_or_default() == "Failed" {
                    error!("Pod which we are awaiting has entered failed state instead of running");
                    return Err(ApplicationError::PodCouldNotStart.into());
                }
                if s.phase.clone().unwrap_or_default() == "Succeeded" {
                    error!("Pod which we are awaiting has completed instead");
                    return Err(ApplicationError::PodDidNotWait.into());
                }
            }
            _ => {}
        }
    }
    Ok(())
}

pub async fn is_pod_running(pods: &Api<Pod>, pod: Pod) -> Result<bool> {
    let got_pod = pods
        .get(
            &pod.metadata
                .name
                .ok_or(anyhow!("pod metadata must have a name"))?,
        )
        .await?;
    let phase = got_pod
        .status
        .ok_or(anyhow!("pod has no status"))?
        .phase
        .ok_or(anyhow!("pod has no status.phase"))?;
    if phase == "Running" {
        Ok(true)
    } else {
        Ok(false)
    }
}

pub async fn is_pod_finished(pods: &Api<Pod>, pod: &Pod) -> Result<bool> {
    let got_pod = pods
        .get(
            &pod.metadata
                .name
                .as_ref()
                .ok_or(anyhow!("pod metadata must have a name"))?,
        )
        .await?;
    let phase = got_pod
        .status
        .ok_or(anyhow!("pod has no status"))?
        .phase
        .ok_or(anyhow!("pod has no status.phase"))?;
    if phase == "Failed" {
        Ok(true)
    } else if phase == "Succeeded" {
        Ok(true)
    } else {
        Ok(false)
    }
}

pub async fn wait_for_pod_running(pods: &Api<Pod>, pod: &Pod) -> Result<()> {
    let (tx, mut rx) = tokio::sync::mpsc::channel(1);
    let xtx = tx.clone();
    let xpods = pods.clone();
    let xpod = pod.clone();
    let _p: JoinHandle<Result<()>> = tokio::spawn(async move {
        let r = is_pod_running(&xpods, xpod).await;
        if let Ok(true) = r {
            xtx.send(Ok(())).await.expect("couldn't send to channel");
        }
        if let Err(e) = r {
            xtx.send(Err(e)).await.expect("couldn't send to channel");
        }
        Ok(())
    });
    let xtx = tx.clone();
    let xpods = pods.clone();
    let xpod = pod.clone();
    let _w = tokio::spawn(async move {
        xtx.send(wait_for_pod_running_watch(&xpods, xpod).await)
            .await
            .expect("couldn't send on channel");
    });
    let r = rx.recv().await;
    if r.is_none() {
        bail!("failed to read API while waiting for pod to start");
    }
    let r = r.expect("failed to extract value after checking for None");
    r
}
